-- md5asm.ex - ASM routine generator for md5.e (version 1.11)
-- Copyright (C) 2000  Davi Tassinari de Figueiredo
--
-- If you wish to contact me, send an e-mail to davitf@usa.net .
--
-- You can get the latest version of this program from my Euphoria page:
-- http://davitf.n3.net/
--
--
-- License terms and disclaimer:
--
-- Permission is granted to anyone to use this software for any purpose,
-- including commercial applications, and to alter it and redistribute it
-- freely, subject to the following restrictions:
--
-- 1. The origin of this software must not be misrepresented; you must not
--    claim that you wrote the original software or remove the original
--    authors' names.
-- 2. Altered source versions must be plainly marked as such, and must not
--    be misrepresented as being the original software.
-- 3. All source distributions, with or without modifications, must be
--    distributed under this license. If this software's source code is
--    distributed as part of a larger product, this item does not apply to
--    the rest of the product.
-- 4. If you use this software in a product, an acknowledgment in the
--    product documentation is required. If the source code for the product
--    is not freely distributed, you must include information on how to
--    freely obtain the original software's source code.
--
-- This software is provided 'as-is', without any express or implied
-- warranty.  In no event will the authors be held liable for any damages
-- arising from the use of this software.
--
-- If you want to distribute this software in a way not allowed by this
-- license, or distribute the source under different license terms, contact
-- the authors for permission.


atom file
file = open("md5asm.asm","w")

constant A = "eax", B = "ebx", C = "ecx", D = "edx",
	 data_add = "esi", temp1 = "ebp", temp2 = "edi"

procedure FF(sequence a, sequence b, sequence c, sequence d, atom x_index,
	     atom rol, atom t_element, atom index)
    printf(file, "; Step %d\n", index)

    -- F (b, c, d) = (b and c) or (not(b) and d)
    printf(file, "mov %s, %s\n", {temp1, b})
    printf(file, "and %s, %s\n", {temp1, c})
    printf(file, "mov %s, %s\n", {temp2, b})
    printf(file, "not %s\n", {temp2})
    printf(file, "and %s, %s\n", {temp2, d})
    printf(file, "or %s, %s\n", {temp1, temp2})

    -- a = b + ((a + F(b,c,d) + X[k] + T[i]) <<< s).
    printf(file, "add %s, %s\n", {a, temp1})  -- a = a + the result of F
    printf(file, "add %s, [%s + %d]\n", {a, data_add, x_index * 4})   -- a = a + x[x_index]
    printf(file, "add %s, #%08x\n", {a, t_element})
    printf(file, "rol %s, %d\n", {a, rol})
    printf(file, "add %s, %s\n", {a, b})

    puts(file, "\n")
end procedure

procedure GG(sequence a, sequence b, sequence c, sequence d, atom x_index,
	     atom rol, atom t_element, atom index)
    printf(file, "; Step %d\n", index)

    -- G (b, c, d) = (b and d) or (c and not(d))
    printf(file, "mov %s, %s\n", {temp1, b})
    printf(file, "and %s, %s\n", {temp1, d})
    printf(file, "mov %s, %s\n", {temp2, d})
    printf(file, "not %s\n", {temp2})
    printf(file, "and %s, %s\n", {temp2, c})
    printf(file, "or %s, %s\n", {temp1, temp2})

    -- a = b + ((a + G(b,c,d) + X[k] + T[i]) <<< s).
    printf(file, "add %s, %s\n", {a, temp1})  -- a = a + the result of G
    printf(file, "add %s, [%s + %d]\n", {a, data_add, x_index * 4})   -- a = a + x[x_index]
    printf(file, "add %s, #%08x\n", {a, t_element})
    printf(file, "rol %s, %d\n", {a, rol})
    printf(file, "add %s, %s\n", {a, b})

    puts(file, "\n")
end procedure

procedure HH(sequence a, sequence b, sequence c, sequence d, atom x_index,
	     atom rol, atom t_element, atom index)
    printf(file, "; Step %d\n", index)

    -- H (b, c, d) = b xor c xor d
    printf(file, "mov %s, %s\n", {temp1, b})
    printf(file, "xor %s, %s\n", {temp1, c})
    printf(file, "xor %s, %s\n", {temp1, d})

    -- a = b + ((a + H(b,c,d) + X[k] + T[i]) <<< s).
    printf(file, "add %s, %s\n", {a, temp1})  -- a = a + the result of H
    printf(file, "add %s, [%s + %d]\n", {a, data_add, x_index * 4})   -- a = a + x[x_index]
    printf(file, "add %s, #%08x\n", {a, t_element})
    printf(file, "rol %s, %d\n", {a, rol})
    printf(file, "add %s, %s\n", {a, b})

    puts(file, "\n")
end procedure

procedure II(sequence a, sequence b, sequence c, sequence d, atom x_index,
	     atom rol, atom t_element, atom index)
    printf(file, "; Step %d\n", index)

    -- I (b, c, d) = c xor (b or not(d))
    printf(file, "mov %s, %s\n", {temp1, d})
    printf(file, "not %s\n", {temp1})
    printf(file, "or %s, %s\n", {temp1, b})
    printf(file, "xor %s, %s\n", {temp1, c})

    -- a = b + ((a + H(b,c,d) + X[k] + T[i]) <<< s).
    printf(file, "add %s, %s\n", {a, temp1})  -- a = a + the result of H
    printf(file, "add %s, [%s + %d]\n", {a, data_add, x_index * 4})   -- a = a + x[x_index]
    printf(file, "add %s, #%08x\n", {a, t_element})
    printf(file, "rol %s, %d\n", {a, rol})
    printf(file, "add %s, %s\n", {a, b})

    puts(file, "\n")
end procedure




constant S11 = 7, S12 = 12, S13 = 17, S14 = 22, S21 = 5, S22 = 9, S23 = 14,
	 S24 = 20, S31 = 4, S32 = 11, S33 = 16, S34 = 23, S41 = 6, S42 = 10,
	 S43 = 15, S44 = 21


puts(file, "; MD5 calculation routine in Assembly\n")
puts(file, "; Copyright (C) 2000 Davi T. Figueiredo\n")
puts(file, "; This routine has been created automatically by md5asm.ex .\n\n")

-- Store old registers
printf(file,"push %s\npush %s\npush %s\npush %s\npush %s\npush %s\npush %s\n\n",{A,B,C,D,data_add,temp1,temp2})

-- Read words
printf(file,"mov %s, [a_address]\nmov %s, [b_address]\nmov %s, [c_address]\nmov %s, [d_address]\n\n", {A,B,C,D})

-- Read data address
printf(file,"mov %s, data_address\n\n", {data_add})


-- The following lines have been adapted from the reference implementation
-- in RFC 1321.

-- Round 1
puts(file, "; Round 1\n\n")
-- Generate code
FF (A, B, C, D,  0, S11, #D76AA478,  1)
FF (D, A, B, C,  1, S12, #E8C7B756,  2)
FF (C, D, A, B,  2, S13, #242070DB,  3)
FF (B, C, D, A,  3, S14, #C1BDCEEE,  4)
FF (A, B, C, D,  4, S11, #F57C0FAF,  5)
FF (D, A, B, C,  5, S12, #4787C62A,  6)
FF (C, D, A, B,  6, S13, #A8304613,  7)
FF (B, C, D, A,  7, S14, #FD469501,  8)
FF (A, B, C, D,  8, S11, #698098D8,  9)
FF (D, A, B, C,  9, S12, #8B44F7AF, 10)
FF (C, D, A, B, 10, S13, #FFFF5BB1, 11)
FF (B, C, D, A, 11, S14, #895CD7BE, 12)
FF (A, B, C, D, 12, S11, #6B901122, 13)
FF (D, A, B, C, 13, S12, #FD987193, 14)
FF (C, D, A, B, 14, S13, #A679438E, 15)
FF (B, C, D, A, 15, S14, #49B40821, 16)

-- Round 2
puts(file, "; Round 2\n\n")
-- Generate code
GG (A, B, C, D,  1, S21, #F61E2562, 17)
GG (D, A, B, C,  6, S22, #C040B340, 18)
GG (C, D, A, B, 11, S23, #265E5A51, 19)
GG (B, C, D, A,  0, S24, #E9B6C7AA, 20)
GG (A, B, C, D,  5, S21, #D62F105D, 21)
GG (D, A, B, C, 10, S22, #02441453, 22)
GG (C, D, A, B, 15, S23, #D8A1E681, 23)
GG (B, C, D, A,  4, S24, #E7D3FBC8, 24)
GG (A, B, C, D,  9, S21, #21E1CDE6, 25)
GG (D, A, B, C, 14, S22, #C33707D6, 26)
GG (C, D, A, B,  3, S23, #F4D50D87, 27)
GG (B, C, D, A,  8, S24, #455A14ED, 28)
GG (A, B, C, D, 13, S21, #A9E3E905, 29)
GG (D, A, B, C,  2, S22, #FCEFA3F8, 30)
GG (C, D, A, B,  7, S23, #676F02D9, 31)
GG (B, C, D, A, 12, S24, #8D2A4C8A, 32)

-- Round 3
puts(file, "; Round 3\n\n")
-- Generate code
HH (A, B, C, D,  5, S31, #FFFA3942, 33)
HH (D, A, B, C,  8, S32, #8771F681, 34)
HH (C, D, A, B, 11, S33, #6D9D6122, 35)
HH (B, C, D, A, 14, S34, #FDE5380C, 36)
HH (A, B, C, D,  1, S31, #A4BEEA44, 37)
HH (D, A, B, C,  4, S32, #4BDECFA9, 38)
HH (C, D, A, B,  7, S33, #F6BB4B60, 39)
HH (B, C, D, A, 10, S34, #BEBFBC70, 40)
HH (A, B, C, D, 13, S31, #289B7EC6, 41)
HH (D, A, B, C,  0, S32, #EAA127FA, 42)
HH (C, D, A, B,  3, S33, #D4EF3085, 43)
HH (B, C, D, A,  6, S34, #04881D05, 44)
HH (A, B, C, D,  9, S31, #D9D4D039, 45)
HH (D, A, B, C, 12, S32, #E6DB99E5, 46)
HH (C, D, A, B, 15, S33, #1FA27CF8, 47)
HH (B, C, D, A,  2, S34, #C4AC5665, 48)

-- Round 4
puts(file, "; Round 4\n\n")
-- Generate code
II (A, B, C, D,  0, S41, #F4292244, 49)
II (D, A, B, C,  7, S42, #432AFF97, 50)
II (C, D, A, B, 14, S43, #AB9423A7, 51)
II (B, C, D, A,  5, S44, #FC93A039, 52)
II (A, B, C, D, 12, S41, #655B59C3, 53)
II (D, A, B, C,  3, S42, #8F0CCC92, 54)
II (C, D, A, B, 10, S43, #FFEFF47D, 55)
II (B, C, D, A,  1, S44, #85845DD1, 56)
II (A, B, C, D,  8, S41, #6FA87E4F, 57)
II (D, A, B, C, 15, S42, #FE2CE6E0, 58)
II (C, D, A, B,  6, S43, #A3014314, 59)
II (B, C, D, A, 13, S44, #4E0811A1, 60)
II (A, B, C, D,  4, S41, #F7537E82, 61)
II (D, A, B, C, 11, S42, #BD3AF235, 62)
II (C, D, A, B,  2, S43, #2AD7D2BB, 63)
II (B, C, D, A,  9, S44, #EB86D391, 64)


-- Update words
printf(file,"add [a_address], %s\nadd [b_address], %s\nadd [c_address], %s\nadd [d_address], %s\n\n", {A, B, C, D})

-- Restore old registers
printf(file,"pop %s\npop %s\npop %s\npop %s\npop %s\npop %s\npop %s\n\nret\n",{ temp2,temp1,data_add,D,C,B,A})

-- End of routine
close(file)

-- I used to use this line while debugging to speed things up
-- system("ex asm md5asm",2)

